@@ -21,10 +21,12 @@ from os_win.utils.compute import livemigrationutils
 from os_win.utils.compute import rdpconsoleutils
 from os_win.utils.compute import vmutils
 from os_win.utils import hostutils
+from os_win.utils.io import namedpipe
 from os_win.utils.network import networkutils
 from os_win.utils import pathutils
 from os_win.utils.storage.initiator import iscsi_cli_utils
 from os_win.utils.storage.initiator import iscsi_wmi_utils
+from os_win.utils.storage import smbutils
 from os_win.utils.storage.target import iscsi_target_utils
 from os_win.utils.storage.virtdisk import vhdutils
 
@@ -88,9 +90,17 @@ def get_livemigrationutils():
     return livemigrationutils.LiveMigrationUtils()
 
 
+def get_smbutils():
+    return smbutils.SMBUtils()
+
+
 def get_rdpconsoleutils():
     return rdpconsoleutils.RDPConsoleUtils()
 
 
 def get_iscsi_target_utils():
     return iscsi_target_utils.ISCSITargetUtils()
+
+
+def get_named_pipe_handler(*args, **kwargs):
+    return namedpipe.NamedPipeHandler(*args, **kwargs)
